/*
 * Copyright (C) 2015 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.google.auto.value.annotations;

import com.google.common.annotations.GwtCompatible;

/**
 * An annotation that is marked {@code @GwtCompatible} and that contains an array member.
 *
 * @author emcmanus@google.com (Éamonn McManus)
 */
@GwtCompatible
public @interface GwtArrays {
  String[] strings();
  int[] ints();
}
